from __future__ import annotations

import asyncio
import inspect
import threading
from asyncio import CancelledError
from concurrent.futures import Future
from contextlib import suppress
from functools import partial
from typing import TYPE_CHECKING, Any, cast

from deprecated.sphinx import versionadded

from coredis._utils import CaseAndEncodingInsensitiveEnum, b, hash_slot, nativestr
from coredis.commands.constants import CommandName
from coredis.connection import BaseConnection, Connection
from coredis.exceptions import ConnectionError, PubSubError, TimeoutError
from coredis.response.types import PubSubMessage
from coredis.retry import (
    CompositeRetryPolicy,
    ConstantRetryPolicy,
    NoRetryPolicy,
    RetryPolicy,
)
from coredis.typing import (
    AnyStr,
    Awaitable,
    Callable,
    Dict,
    Generic,
    List,
    MutableMapping,
    Optional,
    ResponsePrimitive,
    ResponseType,
    StringT,
    TypeVar,
    Union,
    ValueT,
)

if TYPE_CHECKING:
    import coredis.client
    import coredis.connection
    import coredis.pool

T = TypeVar("T")


PoolT = TypeVar("PoolT", bound="coredis.pool.ConnectionPool")

#: Callables for message handler callbacks. The callbacks
#:  can be sync or async.
SubscriptionCallback = Union[
    Callable[[PubSubMessage], Awaitable[None]], Callable[[PubSubMessage], None]
]


class PubSubMessageTypes(CaseAndEncodingInsensitiveEnum):
    MESSAGE = b"message"
    PMESSAGE = b"pmessage"
    SMESSAGE = b"smessage"
    SUBSCRIBE = b"subscribe"
    UNSUBSCRIBE = b"unsubscribe"
    PSUBSCRIBE = b"psubscribe"
    PUNSUBSCRIBE = b"punsubscribe"
    SSUBSCRIBE = b"ssubscribe"
    SUNSUBSCRIBE = b"sunsubscribe"


class BasePubSub(Generic[AnyStr, PoolT]):
    PUBLISH_MESSAGE_TYPES = {
        PubSubMessageTypes.MESSAGE.value,
        PubSubMessageTypes.PMESSAGE.value,
    }
    SUBUNSUB_MESSAGE_TYPES = {
        PubSubMessageTypes.SUBSCRIBE.value,
        PubSubMessageTypes.PSUBSCRIBE.value,
        PubSubMessageTypes.UNSUBSCRIBE.value,
        PubSubMessageTypes.PUNSUBSCRIBE.value,
    }
    UNSUBSCRIBE_MESSAGE_TYPES = {
        PubSubMessageTypes.UNSUBSCRIBE.value,
        PubSubMessageTypes.PUNSUBSCRIBE.value,
    }

    channels: MutableMapping[StringT, Optional[SubscriptionCallback]]
    patterns: MutableMapping[StringT, Optional[SubscriptionCallback]]

    def __init__(
        self,
        connection_pool: PoolT,
        ignore_subscribe_messages: bool = False,
        retry_policy: Optional[RetryPolicy] = CompositeRetryPolicy(
            ConstantRetryPolicy((ConnectionError,), 3, 0.1),
            ConstantRetryPolicy((TimeoutError,), 2, 0.1),
        ),
    ):
        self.connection_pool = connection_pool
        self.ignore_subscribe_messages = ignore_subscribe_messages
        self.connection: Optional[coredis.connection.Connection] = None
        self._retry_policy = retry_policy or NoRetryPolicy()
        self.reset()

    async def _ensure_encoding(self) -> None:
        if hasattr(self, "encoding"):
            return

        await self.connection_pool.initialize()
        conn = await self.connection_pool.get_connection(b"pubsub")

        try:
            self.encoding = conn.encoding
            self.decode_responses = conn.decode_responses
        finally:
            self.connection_pool.release(conn)

    def __del__(self) -> None:
        self.reset()

    def reset(self) -> None:
        """
        Clear subscriptions and disconnect and release any
        connection(s) back to the connection pool.
        """
        if self.connection:
            self.connection.disconnect()
            self.connection.clear_connect_callbacks()
            self.connection_pool.release(self.connection)
            self.connection = None
        self.channels = {}
        self.patterns = {}

    def close(self) -> None:
        self.reset()

    async def reset_connections(self, exc: Optional[BaseException] = None) -> None:
        pass

    async def on_connect(self, connection: BaseConnection) -> None:
        """
        Re-subscribe to any channels and patterns previously subscribed to

        :meta private:
        """

        if self.channels:
            await self.subscribe(
                **{
                    k.decode(self.encoding) if isinstance(k, bytes) else k: v
                    for k, v in self.channels.items()
                }
            )

        if self.patterns:
            await self.psubscribe(
                **{
                    k.decode(self.encoding) if isinstance(k, bytes) else k: v
                    for k, v in self.patterns.items()
                }
            )

    def encode(self, value: StringT) -> StringT:
        """
        Encodes the value so that it's identical to what we'll read off the
        connection

        :meta private:
        """

        if self.decode_responses and isinstance(value, bytes):
            value = nativestr(value, self.encoding)
        elif not self.decode_responses and isinstance(value, str):
            value = b(value, self.encoding)

        return value

    @property
    def subscribed(self) -> bool:
        """Indicates if there are subscriptions to any channels or patterns"""

        return bool(self.channels or self.patterns)

    async def execute_command(
        self, command: bytes, *args: ValueT, **options: ValueT
    ) -> Optional[ResponseType]:
        """
        Executes a publish/subscribe command

        :meta private:
        """

        # NOTE: don't parse the response in this function -- it could pull a
        # legitimate message off the stack if the connection is already
        # subscribed to one or more channels

        await self._ensure_encoding()

        if self.connection is None:
            self.connection = await self.connection_pool.get_connection()
            self.connection.register_connect_callback(self.on_connect)
        assert self.connection
        return await self._execute(
            self.connection, self.connection.send_command, command, *args
        )

    async def _execute(
        self,
        connection: BaseConnection,
        command: Union[
            Callable[..., Awaitable[None]], Callable[..., Awaitable[ResponseType]]
        ],
        *args: ValueT,
    ) -> Optional[ResponseType]:
        try:
            return await command(*args)
        except asyncio.CancelledError:
            # do not retry if coroutine is cancelled
            if await connection.can_read():  # noqa
                connection.disconnect()
            raise

    async def parse_response(
        self, block: bool = True, timeout: Optional[float] = None
    ) -> ResponseType:
        """
        Parses the response from a publish/subscribe command

        :meta private:
        """
        assert self.connection
        coro = self._execute(
            self.connection,
            partial(
                self.connection.fetch_push_message,
                push_message_types=self.SUBUNSUB_MESSAGE_TYPES
                | self.PUBLISH_MESSAGE_TYPES,
            ),
        )

        try:
            return await asyncio.wait_for(
                coro, timeout if (timeout and timeout > 0) else None
            )
        except asyncio.TimeoutError:
            return None

    async def psubscribe(
        self,
        *patterns: StringT,
        **pattern_handlers: Optional[SubscriptionCallback],
    ) -> None:
        """
        Subscribes to channel patterns. Patterns supplied as keyword arguments
        expect a pattern name as the key and a callable as the value. A
        pattern's callable will be invoked automatically when a message is
        received on that pattern rather than producing a message via
        :meth:`listen`.
        """
        await self._ensure_encoding()

        new_patterns: MutableMapping[StringT, Optional[SubscriptionCallback]] = {}
        new_patterns.update(dict.fromkeys(map(self.encode, patterns)))

        for pattern, handler in pattern_handlers.items():
            new_patterns[self.encode(pattern)] = handler
        await self.execute_command(CommandName.PSUBSCRIBE, *new_patterns.keys())
        # update the patterns dict AFTER we send the command. we don't want to
        # subscribe twice to these patterns, once for the command and again
        # for the reconnection.
        self.patterns.update(new_patterns)

    async def punsubscribe(self, *patterns: StringT) -> None:
        """
        Unsubscribes from the supplied patterns. If empy, unsubscribe from
        all patterns.
        """
        await self._ensure_encoding()
        await self.execute_command(CommandName.PUNSUBSCRIBE, *patterns)

    async def subscribe(
        self,
        *channels: StringT,
        **channel_handlers: Optional[SubscriptionCallback],
    ) -> None:
        """
        Subscribes to channels. Channels supplied as keyword arguments expect
        a channel name as the key and a callable as the value. A channel's
        callable will be invoked automatically when a message is received on
        that channel rather than producing a message via :meth:`listen` or
        :meth:`get_message`.
        """

        await self._ensure_encoding()

        new_channels: MutableMapping[StringT, Optional[SubscriptionCallback]] = {}
        new_channels.update(dict.fromkeys(map(self.encode, channels)))

        for channel, handler in channel_handlers.items():
            new_channels[self.encode(channel)] = handler
        await self.execute_command(CommandName.SUBSCRIBE, *new_channels.keys())
        # update the channels dict AFTER we send the command. we don't want to
        # subscribe twice to these channels, once for the command and again
        # for the reconnection.
        self.channels.update(new_channels)

    async def unsubscribe(self, *channels: StringT) -> None:
        """
        Unsubscribes from the supplied channels. If empty, unsubscribe from
        all channels
        """

        await self._ensure_encoding()
        await self.execute_command(CommandName.UNSUBSCRIBE, *channels)

    async def listen(self) -> Optional[PubSubMessage]:
        """
        Listens for messages on channels this client has been subscribed to
        """

        if self.subscribed:
            response = await self.parse_response(block=True)
            if response:
                return await self.handle_message(response)
        return None

    async def get_message(
        self,
        ignore_subscribe_messages: bool = False,
        timeout: Optional[Union[int, float]] = None,
    ) -> Optional[PubSubMessage]:
        """
        Gets the next message if one is available, otherwise None.

        :param ignore_subscribe_messages: Whether to skip subscription
         acknowledgement messages
        :param timeout: Number of seconds to wait for a message to be available
         on the connection. If the ``None`` the command will block forever.
        """

        response = await self._retry_policy.call_with_retries(
            lambda: self.parse_response(block=False, timeout=timeout),
            failure_hook=self.reset_connections,
        )

        if response:
            return await self.handle_message(response, ignore_subscribe_messages)

        return None

    async def handle_message(
        self, response: ResponseType, ignore_subscribe_messages: bool = False
    ) -> Optional[PubSubMessage]:
        """
        Parses a pub/sub message. If the channel or pattern was subscribed to
        with a message handler, the handler is invoked instead of a parsed
        message being returned.

        :meta private:
        """
        r = cast(List[ResponsePrimitive], response)
        message_type = b(r[0])
        message_type_str = nativestr(r[0])
        message: PubSubMessage

        if message_type in self.SUBUNSUB_MESSAGE_TYPES:
            message = PubSubMessage(
                type=message_type_str,
                pattern=cast(StringT, r[1]) if message_type[0] == ord(b"p") else None,
                # This field is populated in all cases for backward compatibility
                # as older versions were incorrectly populating the channel
                # with the pattern on psubscribe/punsubscribe responses.
                channel=cast(StringT, r[1]),
                data=cast(int, r[2]),
            )
        elif message_type in self.PUBLISH_MESSAGE_TYPES:
            if message_type == PubSubMessageTypes.PMESSAGE:
                message = PubSubMessage(
                    type="pmessage",
                    pattern=cast(StringT, r[1]),
                    channel=cast(StringT, r[2]),
                    data=cast(StringT, r[3]),
                )
            else:
                message = PubSubMessage(
                    type="message",
                    pattern=None,
                    channel=cast(StringT, r[1]),
                    data=cast(StringT, r[2]),
                )
        else:
            raise PubSubError(f"Unknown message type {message_type_str}")  # noqa

        # if this is an unsubscribe message, remove it from memory
        if message_type in self.UNSUBSCRIBE_MESSAGE_TYPES:
            if message_type == PubSubMessageTypes.PUNSUBSCRIBE:
                subscribed_dict = self.patterns
            else:
                subscribed_dict = self.channels
            subscribed_dict.pop(message["channel"])

        if message_type in self.PUBLISH_MESSAGE_TYPES:
            handler = None
            if message_type == PubSubMessageTypes.PMESSAGE and message["pattern"]:
                handler = self.patterns.get(message["pattern"], None)
            elif message["channel"]:
                handler = self.channels.get(message["channel"], None)

            if handler:
                handler_response = handler(message)
                if inspect.isawaitable(handler_response):
                    await handler_response
                return None
        else:
            # this is a subscribe/unsubscribe message. ignore if we don't
            # want them

            if ignore_subscribe_messages or self.ignore_subscribe_messages:
                return None

        return message

    def run_in_thread(self, poll_timeout: float = 1.0) -> PubSubWorkerThread:
        """
        Run the listeners in a thread. For each message received on a
        subscribed channel or pattern the registered handlers will be invoked.

        To stop listening invoke :meth:`~coredis.commands.pubsub.PubSubWorkerThread.stop`
        on the returned instance
        """
        for channel, handler in self.channels.items():
            if handler is None:
                raise PubSubError(f"Channel: {channel!r} has no handler registered")

        for pattern, handler in self.patterns.items():
            if handler is None:
                raise PubSubError(f"Pattern: {pattern!r} has no handler registered")
        thread = PubSubWorkerThread(self, poll_timeout=poll_timeout)
        thread.start()

        return thread


class PubSub(BasePubSub[AnyStr, "coredis.pool.ConnectionPool"]):
    """
    Pub/Sub implementation to be used with :class:`coredis.Redis`
    that is returned by :meth:`coredis.Redis.pubsub`
    """


class ClusterPubSub(BasePubSub[AnyStr, "coredis.pool.ClusterConnectionPool"]):
    """
    Pub/Sub implementation to be used with :class:`coredis.RedisCluster`
    that is returned by :meth:`coredis.RedisCluster.pubsub`

    .. note:: This implementation does not particularly benefit from having
       multiple nodes in a cluster as it subscribes to messages sent to channels
       using ``PUBLISH`` which in cluster mode results in the message being
       broadcasted to every node in the cluster. For this reason the subscribing
       client can subscribe to any node in the cluster to receive messages sent to
       any channel - which inherently limits the potential for scaling.

       :redis-version:`7.0` introduces the concept of Sharded Pub/Sub which
       can be accessed by instead using :meth:`coredis.RedisCluster.sharded_pubsub`
       which uses the implementation in :class:`coredis.commands.ShardedPubSub`.


    """

    async def execute_command(
        self, command: bytes, *args: ValueT, **options: ValueT
    ) -> Optional[ResponseType]:
        await self.connection_pool.initialize()

        if self.connection is None:
            await self.reset_connections(None)

        assert self.connection
        return await self._execute(
            self.connection, self.connection.send_command, command, *args
        )

    async def reset_connections(self, exc: Optional[BaseException] = None) -> None:
        if self.connection:
            self.connection.disconnect()
            self.connection_pool.initialized = False

        await self.connection_pool.initialize()

        self.connection = await self.connection_pool.get_connection(b"pubsub")
        self.connection.register_connect_callback(self.on_connect)


@versionadded(version="3.6.0")
class ShardedPubSub(BasePubSub[AnyStr, "coredis.pool.ClusterConnectionPool"]):
    """
    Sharded Pub/Sub implementation to be used with :class:`coredis.RedisCluster`
    that is returned by :meth:`coredis.RedisCluster.sharded_pubsub`

    For details about the server architecture refer to the `Redis manual entry
    on Sharded Pub/sub <https://redis.io/docs/manual/pubsub/#sharded-pubsub>`__.

    New in :redis-version:`7.0.0`

    .. warning:: Sharded PubSub only supports subscription by channel and does
       **NOT** support pattern based subscriptions.
    """

    PUBLISH_MESSAGE_TYPES = {
        PubSubMessageTypes.MESSAGE.value,
        PubSubMessageTypes.SMESSAGE.value,
    }
    SUBUNSUB_MESSAGE_TYPES = {
        PubSubMessageTypes.SSUBSCRIBE.value,
        PubSubMessageTypes.SUNSUBSCRIBE.value,
    }
    UNSUBSCRIBE_MESSAGE_TYPES = {PubSubMessageTypes.SUNSUBSCRIBE.value}

    def __init__(
        self,
        connection_pool: coredis.pool.ClusterConnectionPool,
        ignore_subscribe_messages: bool = False,
        retry_policy: Optional[RetryPolicy] = None,
        read_from_replicas: bool = False,
    ):
        self.shard_connections: Dict[str, Connection] = {}
        self.channel_connection_mapping: Dict[StringT, Connection] = {}
        self.pending_tasks: Dict[str, asyncio.Task[ResponseType]] = {}
        self.read_from_replicas = read_from_replicas
        super().__init__(connection_pool, ignore_subscribe_messages, retry_policy)

    async def subscribe(
        self,
        *channels: StringT,
        **channel_handlers: Optional[SubscriptionCallback],
    ) -> None:
        """
        :param channels: The shard channels to subscribe to.
        :param channel_handlers: Channels supplied as keyword arguments expect
         a channel name as the key and a callable as the value. A channel's
         callable will be invoked automatically when a message is received on
         that channel rather than producing a message via :meth:`listen` or
         :meth:`get_message`.
        """

        await self._ensure_encoding()
        new_channels: MutableMapping[StringT, Optional[SubscriptionCallback]] = {}
        new_channels.update(dict.fromkeys(map(self.encode, channels)))

        for channel, handler in channel_handlers.items():
            new_channels[self.encode(channel)] = handler
        for new_channel in new_channels.keys():
            await self.execute_command(
                CommandName.SSUBSCRIBE, new_channel, sharded=True
            )
        self.channels.update(new_channels)

    async def unsubscribe(self, *channels: StringT) -> None:
        """
        :param channels: The shard channels to unsubscribe from. If None are provided,
         this will effectively unsubscribe the client from all channels
         previously subscribed to.
        """

        await self._ensure_encoding()
        for channel in channels or list(self.channels.keys()):
            await self.execute_command(CommandName.SUNSUBSCRIBE, channel, sharded=True)

    async def psubscribe(
        self,
        *patterns: StringT,
        **pattern_handlers: Optional[SubscriptionCallback],
    ) -> None:
        """
        Not available in sharded pubsub

        :meta private:
        """
        raise NotImplementedError(
            "Sharded PubSub does not support subscription by pattern"
        )

    async def punsubscribe(self, *patterns: StringT) -> None:
        """
        Not available in sharded pubsub

        :meta private:
        """
        raise NotImplementedError(
            "Sharded PubSub does not support subscription by pattern"
        )

    async def execute_command(
        self, command: bytes, *args: ValueT, **options: ValueT
    ) -> Optional[ResponseType]:
        await self.connection_pool.initialize()

        assert isinstance(args[0], (bytes, str))
        channel = nativestr(args[0])
        slot = hash_slot(b(channel))
        node = self.connection_pool.nodes.node_from_slot(slot)
        if node and node.node_id:
            key = node.node_id
            if self.shard_connections.get(key) is None:
                self.shard_connections[key] = await self.connection_pool.get_connection(
                    b"pubsub",
                    channel=channel,
                    node_type="replica" if self.read_from_replicas else "primary",
                )
                # register a callback that re-subscribes to any channels we
                # were listening to when we were disconnected
                self.shard_connections[key].register_connect_callback(self.on_connect)

            self.channel_connection_mapping[args[0]] = self.shard_connections[key]
            assert self.shard_connections[key]
            return await self._execute(
                self.shard_connections[key],
                self.shard_connections[key].send_command,
                command,
                *args,
            )
        raise PubSubError(f"Unable to determine shard for channel {args[0]!r}")

    async def reset_connections(self, exc: Optional[BaseException] = None) -> None:
        for connection in self.shard_connections.values():
            connection.disconnect()
            connection.clear_connect_callbacks()
            self.connection_pool.release(connection)
        self.shard_connections.clear()
        for _, task in self.pending_tasks.items():
            if not task.done():
                task.cancel()
                with suppress(CancelledError):
                    await task
        self.pending_tasks.clear()
        self.connection_pool.disconnect()
        self.connection_pool.reset()
        self.connection_pool.initialized = False
        await self.connection_pool.initialize()
        for channel in self.channels:
            slot = hash_slot(b(channel))
            node = self.connection_pool.nodes.node_from_slot(slot)
            if node and node.node_id:
                key = node.node_id
                self.shard_connections[key] = await self.connection_pool.get_connection(
                    b"pubsub",
                    channel=channel,
                    node_type="replica" if self.read_from_replicas else "primary",
                )
                # register a callback that re-subscribes to any channels we
                # were listening to when we were disconnected
                self.shard_connections[key].register_connect_callback(self.on_connect)
                self.channel_connection_mapping[channel] = self.shard_connections[key]

    async def parse_response(
        self, block: bool = True, timeout: Optional[float] = None
    ) -> ResponseType:
        if not self.shard_connections:
            raise RuntimeError(
                "pubsub connection not set: "
                "did you forget to call subscribe() or psubscribe()?"
            )
        result = None
        # Check any stashed results first.
        if self.pending_tasks:
            for node_id, task in list(self.pending_tasks.items()):
                self.pending_tasks.pop(node_id)
                if task.done():
                    result = task.result()
                    break
                else:
                    done, pending = await asyncio.wait(
                        [task],
                        timeout=0.001,
                        return_when=asyncio.FIRST_COMPLETED,
                    )
                    if done:
                        result = done.pop().result()
                        break
                    else:
                        task.cancel()
                        with suppress(CancelledError):
                            await task
        # If there were no pending results check the shards
        if not result:
            broken_connections = [
                c for c in self.shard_connections.values() if not c.is_connected
            ]
            if broken_connections:
                for connection in broken_connections:
                    try:
                        await connection.connect()
                    except:  # noqa
                        raise ConnectionError("Shard connections not stable")
            tasks: Dict[str, asyncio.Task[ResponseType]] = {
                node_id: asyncio.create_task(
                    self._execute(
                        connection,
                        partial(
                            connection.fetch_push_message,
                            push_message_types=self.SUBUNSUB_MESSAGE_TYPES
                            | self.PUBLISH_MESSAGE_TYPES,
                        ),
                    )
                )
                for node_id, connection in self.shard_connections.items()
                if node_id not in self.pending_tasks
            }
            if tasks:
                done, pending = await asyncio.wait(
                    tasks.values(),
                    timeout=timeout if (timeout and timeout > 0) else None,
                    return_when=asyncio.FIRST_COMPLETED,
                )
                if done:
                    done_task = done.pop()
                    result = done_task.result()

                # Stash any other tasks for the next iteration
                for task in list(done) + list(pending):
                    for node_id, scheduled in tasks.items():
                        if task == scheduled:
                            self.pending_tasks[node_id] = task
        return result

    async def on_connect(self, connection: BaseConnection) -> None:
        """
        Re-subscribe to any channels previously subscribed to

        :meta private:
        """
        for channel, handler in self.channels.items():
            if self.channel_connection_mapping[channel] == connection:
                await self.subscribe(
                    **{
                        (
                            channel.decode(self.encoding)
                            if isinstance(channel, bytes)
                            else channel
                        ): handler
                    }
                )

    def reset(self) -> None:
        for connection in self.shard_connections.values():
            connection.disconnect()
            connection.clear_connect_callbacks()
            self.connection_pool.release(connection)
        for _, task in self.pending_tasks.items():
            task.cancel()
        self.pending_tasks.clear()
        self.shard_connections.clear()
        self.channels = {}
        self.patterns = {}


class PubSubWorkerThread(threading.Thread):
    def __init__(
        self,
        pubsub: BasePubSub[Any, Any],
        poll_timeout: float = 1.0,
    ):
        super().__init__()
        self._pubsub = pubsub
        self._poll_timeout = poll_timeout
        self._running = False
        self._loop = asyncio.get_running_loop()
        self._future: Optional[Future[None]] = None

    async def _run(self) -> None:
        pubsub = self._pubsub
        try:
            while pubsub.subscribed:
                await pubsub.get_message(
                    ignore_subscribe_messages=True, timeout=self._poll_timeout
                )
        except CancelledError:
            await asyncio.gather(pubsub.unsubscribe(), pubsub.punsubscribe())
            pubsub.close()
            self._running = False

    def run(self) -> None:
        """
        :meta private:
        """
        if self._running:
            return
        self._running = True
        self._future = asyncio.run_coroutine_threadsafe(self._run(), self._loop)

    def stop(self) -> None:
        """
        Stop the worker thread from processing any more messages
        """
        if self._future:
            self._future.cancel()



from __future__ import annotations

import asyncio
import os
import threading
import time
import warnings
from itertools import chain
from ssl import SSLContext, VerifyMode
from typing import Any, cast
from urllib.parse import parse_qs, unquote, urlparse

import async_timeout

from coredis.connection import (
    BaseConnection,
    Connection,
    RedisSSLContext,
    UnixDomainSocketConnection,
)
from coredis.exceptions import ConnectionError
from coredis.typing import (
    Callable,
    ClassVar,
    Dict,
    List,
    Optional,
    Set,
    StringT,
    Type,
    TypeVar,
    Union,
    ValueT,
)

FALSE_STRINGS = ("0", "F", "FALSE", "N", "NO")


def to_bool(value: Optional[StringT]) -> Optional[bool]:
    if value is None or value == "":
        return None

    if isinstance(value, str) and value.upper() in FALSE_STRINGS:
        return False

    return bool(value)


_CPT = TypeVar("_CPT", bound="ConnectionPool")


class ConnectionPool:
    """Generic connection pool"""

    #: Mapping of querystring arguments to their parser functions
    URL_QUERY_ARGUMENT_PARSERS: ClassVar[
        Dict[str, Callable[..., Optional[Union[int, float, bool, str]]]]
    ] = {
        "client_name": str,
        "stream_timeout": float,
        "connect_timeout": float,
        "max_connections": int,
        "max_idle_time": int,
        "protocol_version": int,
        "idle_check_interval": int,
        "noreply": bool,
        "noevict": bool,
        "notouch": bool,
    }

    @classmethod
    def from_url(
        cls: Type[_CPT],
        url: str,
        db: Optional[int] = None,
        decode_components: bool = False,
        **kwargs: Any,
    ) -> _CPT:
        """
        Returns a connection pool configured from the given URL.

        For example:

        - ``redis://[:password]@localhost:6379/0``
        - ``rediss://[:password]@localhost:6379/0``
        - ``unix://[:password]@/path/to/socket.sock?db=0``

        Three URL schemes are supported:

        - `redis:// <http://www.iana.org/assignments/uri-schemes/prov/redis>`__
          creates a normal TCP socket connection
        - `rediss:// <http://www.iana.org/assignments/uri-schemes/prov/rediss>`__
          creates a SSL wrapped TCP socket connection
        - ``unix://`` creates a Unix Domain Socket connection

        There are several ways to specify a database number. The parse function
        will return the first specified option:

        - A ``db`` querystring option, e.g. ``redis://localhost?db=0``
        - If using the ``redis://`` scheme, the path argument of the url, e.g.
          ``redis://localhost/0``
        - The ``db`` argument to this function.

        If none of these options are specified, ``db=0`` is used.

        The :paramref:`decode_components` argument allows this function to work with
        percent-encoded URLs. If this argument is set to ``True`` all ``%xx``
        escapes will be replaced by their single-character equivalents after
        the URL has been parsed. This only applies to the ``hostname``,
        ``path``, and ``password`` components. See :attr:`URL_QUERY_ARGUMENT_PARSERS`
        for a comprehensive mapping of querystring parameters to how they are parsed.

        Any additional querystring arguments and keyword arguments will be
        passed along to the class constructor. The querystring
        arguments ``connect_timeout`` and ``stream_timeout`` if supplied
        are parsed as float values.

        .. note:: In the case of conflicting arguments, querystring arguments always win.
        """
        parsed_url = urlparse(url)
        qs = parsed_url.query

        url_options: Dict[
            str,
            Optional[Union[int, float, bool, str, Type[BaseConnection], SSLContext]],
        ] = {}
        for name, value in iter(parse_qs(qs).items()):
            if value and len(value) > 0:
                parser = cls.URL_QUERY_ARGUMENT_PARSERS.get(name)

                if parser:
                    try:
                        url_options[name] = parser(value[0])
                    except (TypeError, ValueError):
                        warnings.warn(
                            UserWarning(
                                "Invalid value for `%s` in connection URL." % name
                            )
                        )
                else:
                    url_options[name] = value[0]

        username: Optional[str] = parsed_url.username
        password: Optional[str] = parsed_url.password
        path: Optional[str] = parsed_url.path
        hostname: Optional[str] = parsed_url.hostname

        if decode_components:
            username = unquote(username) if username else None
            password = unquote(password) if password else None
            path = unquote(path) if path else None
            hostname = unquote(hostname) if hostname else None

        # We only support redis:// and unix:// schemes.

        if parsed_url.scheme == "unix":
            url_options.update(
                {
                    "username": username,
                    "password": password,
                    "path": path,
                    "connection_class": UnixDomainSocketConnection,
                }
            )

        else:
            url_options.update(
                {
                    "host": hostname,
                    "port": int(parsed_url.port or 6379),
                    "username": username,
                    "password": password,
                }
            )

            # If there's a path argument, use it as the db argument if a
            # querystring value wasn't specified

            if "db" not in url_options and path:
                try:
                    url_options["db"] = int(path.replace("/", ""))
                except (AttributeError, ValueError):
                    pass

            if parsed_url.scheme == "rediss":
                keyfile = cast(Optional[str], url_options.pop("ssl_keyfile", None))
                certfile = cast(Optional[str], url_options.pop("ssl_certfile", None))
                check_hostname = cast(
                    Optional[bool], url_options.pop("ssl_check_hostname", None)
                )
                cert_reqs = cast(
                    Optional[Union[str, VerifyMode]],
                    url_options.pop("ssl_cert_reqs", None),
                )
                ca_certs = cast(Optional[str], url_options.pop("ssl_ca_certs", None))
                url_options["ssl_context"] = RedisSSLContext(
                    keyfile, certfile, cert_reqs, ca_certs, check_hostname
                ).get()

        # last shot at the db value
        _db = url_options.get("db", db or 0)
        assert isinstance(_db, (int, str, bytes))
        url_options["db"] = int(_db)

        # update the arguments from the URL values
        kwargs.update(url_options)

        return cls(**kwargs)

    def __init__(
        self,
        *,
        connection_class: Optional[Type[Connection]] = None,
        max_connections: Optional[int] = None,
        max_idle_time: int = 0,
        idle_check_interval: int = 1,
        **connection_kwargs: Optional[Any],
    ) -> None:
        """
        Creates a connection pool. If :paramref:`max_connections` is set, then this
        object raises :class:`~coredis.ConnectionError` when the pool's limit is reached.

        By default, TCP connections are created :paramref:`connection_class` is specified.
        Use :class:`~coredis.UnixDomainSocketConnection` for unix sockets.

        Any additional keyword arguments are passed to the constructor of
        connection_class.
        """
        self.connection_class = connection_class or Connection
        self.connection_kwargs = connection_kwargs
        self.max_connections = max_connections or 2**31
        self.max_idle_time = max_idle_time
        self.idle_check_interval = idle_check_interval
        self.initialized = False
        self.reset()

    async def initialize(self) -> None:
        self.initialized = True

    def __repr__(self) -> str:
        return "{}<{}>".format(
            type(self).__name__, self.connection_class.describe(self.connection_kwargs)
        )

    def __del__(self) -> None:
        self.disconnect()

    async def disconnect_on_idle_time_exceeded(self, connection: Connection) -> None:
        while True:
            if (
                time.time() - connection.last_active_at > self.max_idle_time
                and not connection.requests_pending
            ):
                connection.disconnect()
                if connection in self._available_connections:
                    self._available_connections.remove(connection)
                self._created_connections -= 1
                break
            await asyncio.sleep(self.idle_check_interval)

    def reset(self) -> None:
        self.pid = os.getpid()
        self._created_connections = 0
        self._available_connections: List[Connection] = []
        self._in_use_connections: Set[Connection] = set()
        self._check_lock = threading.Lock()

    def checkpid(self) -> None:  # noqa
        if self.pid != os.getpid():
            with self._check_lock:
                # Double check
                if self.pid == os.getpid():
                    return
                self.disconnect()
                self.reset()

    def peek_available(self) -> Optional[BaseConnection]:
        return self._available_connections[0] if self._available_connections else None

    async def get_connection(
        self,
        command_name: Optional[bytes] = None,
        *args: ValueT,
        acquire: bool = True,
        **kwargs: Optional[ValueT],
    ) -> Connection:
        """Gets a connection from the pool"""
        self.checkpid()
        try:
            connection = self._available_connections.pop()
            if connection.is_connected and connection.needs_handshake:
                await connection.perform_handshake()
        except IndexError:
            if self._created_connections >= self.max_connections:
                raise ConnectionError("Too many connections")
            connection = self._make_connection(**kwargs)

        if acquire:
            self._in_use_connections.add(connection)
        else:
            self._available_connections.append(connection)

        return connection

    def release(self, connection: Connection) -> None:
        """
        Releases the :paramref:`connection` back to the pool
        """
        self.checkpid()

        if connection.pid == self.pid:
            self._in_use_connections.remove(connection)
            self._available_connections.append(connection)

    def disconnect(self) -> None:
        """Closes all connections in the pool"""
        all_conns = chain(self._available_connections, self._in_use_connections)

        for connection in all_conns:
            connection.disconnect()
            self._created_connections -= 1

    def _make_connection(self, **options: Optional[ValueT]) -> Connection:
        """
        Creates a new connection
        """

        self._created_connections += 1
        connection = self.connection_class(
            **self.connection_kwargs,  # type: ignore
        )

        if self.max_idle_time > self.idle_check_interval > 0:
            # do not await the future
            asyncio.ensure_future(self.disconnect_on_idle_time_exceeded(connection))

        return connection


class BlockingConnectionPool(ConnectionPool):
    """
    Blocking connection pool::

        >>> from coredis import Redis
        >>> client = Redis(connection_pool=BlockingConnectionPool())

    It performs the same function as the default
    :class:`~coredis.ConnectionPool`, in that, it maintains a pool of reusable
    connections that can be shared by multiple redis clients.

    The difference is that, in the event that a client tries to get a
    connection from the pool when all of the connections are in use, rather than
    raising a :exc:`~coredis.ConnectionError` (as the default
    :class:`~coredis.ConnectionPool` implementation does), it
    makes the client blocks for a specified number of seconds until
    a connection becomes available.

    Use :paramref:`max_connections` to increase / decrease the pool size::

        >>> pool = BlockingConnectionPool(max_connections=10)

    Use :paramref:`timeout` to tell it either how many seconds to wait for a
    connection to become available, or to block forever::

        >>> # Block forever.
        >>> pool = BlockingConnectionPool(timeout=None)
        >>> # Raise a ``ConnectionError`` after five seconds if a connection is
        >>> # not available.
        >>> pool = BlockingConnectionPool(timeout=5)
    """

    def __init__(
        self,
        connection_class: Optional[Type[Connection]] = None,
        queue_class: Type[asyncio.Queue[Optional[Connection]]] = asyncio.LifoQueue,
        max_connections: Optional[int] = None,
        timeout: int = 20,
        max_idle_time: int = 0,
        idle_check_interval: int = 1,
        **connection_kwargs: Optional[ValueT],
    ):
        self.timeout = timeout
        self.queue_class = queue_class
        self.total_wait = 0
        self.total_allocated = 0
        max_connections = max_connections or 50

        super().__init__(
            connection_class=connection_class or Connection,
            max_connections=max_connections,
            max_idle_time=max_idle_time,
            idle_check_interval=idle_check_interval,
            **connection_kwargs,
        )

    async def disconnect_on_idle_time_exceeded(self, connection: Connection) -> None:
        while True:
            if time.time() - connection.last_active_at > self.max_idle_time:
                # Unlike the non blocking pool, we don't free the connection object,
                # but always reuse it
                connection.disconnect()

                break
            await asyncio.sleep(self.idle_check_interval)

    def reset(self) -> None:
        self._pool: asyncio.Queue[Optional[Connection]] = self.queue_class(
            self.max_connections
        )

        while True:
            try:
                self._pool.put_nowait(None)
            except asyncio.QueueFull:
                break

        super().reset()

    def peek_available(self) -> Optional[BaseConnection]:
        return (
            self._pool._queue[-1]  # type: ignore
            if (self._pool and not self._pool.empty())
            else None
        )

    async def get_connection(
        self,
        command_name: Optional[bytes] = None,
        *args: ValueT,
        acquire: bool = True,
        **kwargs: Optional[ValueT],
    ) -> Connection:
        """Gets a connection from the pool"""
        self.checkpid()

        try:
            async with async_timeout.timeout(self.timeout):
                connection = await self._pool.get()
            if connection and connection.is_connected and connection.needs_handshake:
                await connection.perform_handshake()
        except asyncio.TimeoutError:
            raise ConnectionError("No connection available.")
        if connection is None:
            connection = self._make_connection()

        if acquire:
            self._in_use_connections.add(connection)
        else:
            self._pool.put_nowait(connection)

        return connection

    def release(self, connection: Connection) -> None:
        """Releases the connection back to the pool"""
        _connection: Optional[Connection] = connection

        self.checkpid()

        if _connection and _connection.pid == self.pid:
            self._in_use_connections.remove(_connection)
            try:
                self._pool.put_nowait(_connection)
            except asyncio.QueueFull:
                _connection.disconnect()

    def disconnect(self) -> None:
        """Closes all connections in the pool"""
        pooled_connections: List[Optional[Connection]] = []

        while True:
            try:
                pooled_connections.append(self._pool.get_nowait())
            except asyncio.QueueEmpty:
                break
        for conn in pooled_connections:
            self._pool.put_nowait(conn)

        all_conns = chain(pooled_connections, self._in_use_connections)

        for connection in all_conns:
            if connection is not None:
                connection.disconnect()



from __future__ import annotations

from .basic import BlockingConnectionPool, ConnectionPool
from .cluster import BlockingClusterConnectionPool, ClusterConnectionPool

__all__ = [
    "ConnectionPool",
    "BlockingConnectionPool",
    "ClusterConnectionPool",
    "BlockingClusterConnectionPool",
]



from __future__ import annotations

from collections import ChainMap, OrderedDict
from functools import partial

from coredis._json import json
from coredis._utils import EncodingInsensitiveDict
from coredis.modules.response.types import (
    SearchAggregationResult,
    SearchDocument,
    SearchResult,
)
from coredis.response._callbacks import ResponseCallback
from coredis.response._utils import flat_pairs_to_dict
from coredis.typing import (
    AnyStr,
    Dict,
    List,
    Optional,
    ResponsePrimitive,
    ResponseType,
    StringT,
    TypedDict,
    Union,
    ValueT,
)


class SearchConfigCallback(
    ResponseCallback[
        List[List[ResponsePrimitive]],
        Union[Dict[AnyStr, ResponseType], List[List[ResponsePrimitive]]],
        Dict[AnyStr, ResponsePrimitive],
    ]
):
    def transform(
        self, response: List[List[ResponsePrimitive]], **options: Optional[ValueT]
    ) -> Dict[AnyStr, ResponsePrimitive]:
        pieces = []
        for item in response:
            try:
                v = (item[0], json.loads(item[1]))
            except (ValueError, TypeError):
                v = item
            pieces.append(v)
        return dict(pieces)

    def transform_3(
        self,
        response: Union[Dict[AnyStr, ResponseType], List[List[ResponsePrimitive]]],
        **options: Optional[ValueT],
    ) -> Dict[AnyStr, ResponsePrimitive]:
        if isinstance(response, list):
            return self.transform(response, **options)
        else:
            config = {}
            for item, value in response.items():
                try:
                    config[item] = json.loads(value)
                except (ValueError, TypeError):
                    config[item] = value
            return config


class SearchResultCallback(
    ResponseCallback[
        List[ResponseType],
        Union[List[ResponseType], Dict[AnyStr, ResponseType]],
        SearchResult[AnyStr],
    ]
):
    def transform(
        self, response: List[ResponseType], **options: Optional[ValueT]
    ) -> SearchResult[AnyStr]:
        if options.get("nocontent"):
            return SearchResult[AnyStr](
                response[0],
                tuple(
                    SearchDocument(i, None, None, None, None, {}) for i in response[1:]
                ),
            )
        step = 2
        results = []
        score_idx = payload_idx = sort_key_idx = 0
        if options.get("withscores"):
            score_idx = 1
            step += 1
        if options.get("withpayloads"):
            payload_idx = score_idx + 1
            step += 1
        if options.get("withsortkeys"):
            sort_key_idx = payload_idx + 1
            step += 1

        for k in range(1, len(response) - 1, step):
            section = response[k : k + step]
            score_explain = None
            if options.get("explainscore"):
                score = section[score_idx][0]
                score_explain = section[score_idx][1]
            else:
                score = section[score_idx] if score_idx else None
            fields = EncodingInsensitiveDict(flat_pairs_to_dict(section[-1]))
            if "$" in fields:
                fields = json.loads(fields.pop("$"))
            results.append(
                SearchDocument(
                    section[0],
                    float(score) if score else None,
                    score_explain,
                    section[payload_idx] if payload_idx else None,
                    section[sort_key_idx] if sort_key_idx else None,
                    fields,
                )
            )
        return SearchResult[AnyStr](response[0], tuple(results))

    def transform_3(
        self,
        response: Union[List[ResponseType], Dict[AnyStr, ResponseType]],
        **options: Optional[ValueT],
    ) -> SearchResult[AnyStr]:
        results = []
        if isinstance(response, list):
            return self.transform(response, **options)
        else:
            response = EncodingInsensitiveDict(response)
            for result in response["results"]:
                result = EncodingInsensitiveDict(result)
                score_explain = None
                if options.get("explainscore"):
                    score = result["score"][0]
                    score_explain = result["score"][1]
                else:
                    score = result["score"]
                fields = EncodingInsensitiveDict(result["extra_attributes"])
                if "$" in fields:
                    fields = json.loads(fields.pop("$"))
                results.append(
                    SearchDocument(
                        result["id"],
                        float(score) if score else None,
                        score_explain,
                        result["payload"] if options.get("withpayloads") else None,
                        result["sortkey"] if options.get("withsortkeys") else None,
                        fields,
                    )
                )
            return SearchResult[AnyStr](response["total_results"], tuple(results))


class AggregationResultCallback(
    ResponseCallback[
        List[ResponseType],
        Union[Dict[AnyStr, ResponseType], List[ResponseType]],
        SearchAggregationResult[AnyStr],
    ]
):
    def transform(
        self, response: List[ResponseType], **options: Optional[ValueT]
    ) -> SearchAggregationResult:
        return SearchAggregationResult[AnyStr](
            [
                flat_pairs_to_dict(k, partial(self.try_json, options))
                for k in (
                    response[1:] if not options.get("with_cursor") else response[0][1:]
                )
            ],
            response[1] if options.get("with_cursor") else None,
        )

    def transform_3(
        self,
        response: Union[Dict[AnyStr, ResponseType], List[ResponseType]],
        **options: Optional[ValueT],
    ) -> SearchAggregationResult:
        if (
            options.get("with_cursor")
            and isinstance(response[0], dict)
            or isinstance(response, dict)
        ):
            response, cursor = (
                response if options.get("with_cursor") else (response, None)
            )
            response = EncodingInsensitiveDict(response)
            return SearchAggregationResult[AnyStr](
                [
                    {
                        k: self.try_json(options, v)
                        for k, v in k["extra_attributes"].items()
                    }
                    for k in (response["results"])
                ],
                cursor,
            )
        else:
            return self.transform(response, **options)

    @staticmethod
    def try_json(options, value):
        if not options.get("dialect", None) == 3:
            return value
        try:
            return json.loads(value)
        except ValueError:
            return value


class SpellCheckResult(TypedDict):
    term: StringT
    suggestions: OrderedDict[StringT, int]


class SpellCheckCallback(
    ResponseCallback[
        List[ResponseType],
        Union[Dict[AnyStr, ResponseType], List[ResponseType]],
        SpellCheckResult,
    ]
):
    def transform(
        self, response: List[ResponseType], **options: Optional[ValueT]
    ) -> SpellCheckResult:
        suggestions = {}
        for result in response:
            suggestions[result[1]] = OrderedDict((k[1], float(k[0])) for k in result[2])

        return suggestions

    def transform_3(
        self,
        response: Union[Dict[AnyStr, ResponseType], List[ResponseType]],
        **options: Optional[ValueT],
    ) -> SpellCheckResult:
        if isinstance(response, list):
            return self.transform(response, **options)
        else:
            return {
                key: OrderedDict(ChainMap(*result))
                for key, result in response["results"].items()
            }



from __future__ import annotations

import itertools
from datetime import datetime, timedelta
from typing import List

from deprecated.sphinx import versionadded

from coredis.typing import (
    AnyStr,
    CommandArgList,
    Dict,
    KeyT,
    Literal,
    Mapping,
    Optional,
    Parameters,
    ResponseType,
    Set,
    StringT,
    Tuple,
    Union,
    ValueT,
)

from .._utils import dict_to_flat_list
from ..commands._utils import normalized_milliseconds, normalized_time_milliseconds
from ..commands._validators import (
    mutually_exclusive_parameters,
    mutually_inclusive_parameters,
)
from ..commands._wrappers import CacheConfig, ClusterCommandConfig
from ..commands.constants import CommandFlag, CommandGroup, CommandName, NodeFlag
from ..response._callbacks import (
    ClusterMergeSets,
    IntCallback,
    SetCallback,
    SimpleStringCallback,
    TupleCallback,
)
from ..tokens import PrefixToken, PureToken
from .base import Module, ModuleGroup, module_command
from .response._callbacks.timeseries import (
    ClusterMergeTimeSeries,
    SampleCallback,
    SamplesCallback,
    TimeSeriesCallback,
    TimeSeriesInfoCallback,
    TimeSeriesMultiCallback,
)


def normalized_timestamp(ts: Union[int, datetime, StringT]) -> Union[StringT, int]:
    if isinstance(ts, (bytes, str)):
        return ts
    return normalized_time_milliseconds(ts)


class RedisTimeSeries(Module[AnyStr]):
    NAME = "timeseries"
    FULL_NAME = "RedisTimeSeries"
    DESCRIPTION = """RedisTimeSeries is a Redis module that implements a time series 
data structure. It is designed to be used as a database for time series data, 
and is optimized for fast insertion and retrieval of time series data.
    """
    DOCUMENTATION_URL = "https://redis.io/docs/stack/timeseries/"


@versionadded(version="4.12")
class TimeSeries(ModuleGroup[AnyStr]):
    MODULE = RedisTimeSeries
    COMMAND_GROUP = CommandGroup.TIMESERIES

    @module_command(
        CommandName.TS_CREATE,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def create(
        self,
        key: KeyT,
        retention: Optional[Union[int, timedelta]] = None,
        encoding: Optional[
            Literal[PureToken.COMPRESSED, PureToken.UNCOMPRESSED]
        ] = None,
        chunk_size: Optional[int] = None,
        duplicate_policy: Optional[
            Literal[
                PureToken.BLOCK,
                PureToken.FIRST,
                PureToken.LAST,
                PureToken.MAX,
                PureToken.MIN,
                PureToken.SUM,
            ]
        ] = None,
        labels: Optional[Mapping[StringT, ValueT]] = None,
    ) -> bool:
        """
        Create a new time series with the given key.

        :param key: The key name for the time series.
        :param retention: Maximum age for samples compared to the highest reported timestamp,
         in milliseconds.
        :param encoding: Specifies the series samples encoding format as ``COMPRESSED`` or
         ``UNCOMPRESSED``.
        :param chunk_size: Initial allocation size, in bytes, for the data part of each new chunk.
        :param duplicate_policy: Policy for handling insertion of multiple samples with identical
         timestamps.
        :param labels: A dictionary of labels to be associated with the time series.
        :return: True if the time series was created successfully, False otherwise.
        """
        pieces: CommandArgList = [key]
        if retention is not None:
            pieces.extend([PrefixToken.RETENTION, normalized_milliseconds(retention)])
        if encoding:
            pieces.extend([PrefixToken.ENCODING, encoding])
        if chunk_size is not None:
            pieces.extend([PrefixToken.CHUNK_SIZE, chunk_size])
        if duplicate_policy is not None:
            pieces.extend([PrefixToken.DUPLICATE_POLICY, duplicate_policy])
        if labels:
            pieces.extend(
                [
                    PrefixToken.LABELS,
                    *dict_to_flat_list(labels),  # type: ignore
                ]
            )
        return await self.execute_module_command(
            CommandName.TS_CREATE, *pieces, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.TS_DEL,
        group=COMMAND_GROUP,
        version_introduced="1.6.0",
        module=MODULE,
    )
    async def delete(
        self,
        key: KeyT,
        fromtimestamp: Union[int, datetime, StringT],
        totimestamp: Union[int, datetime, StringT],
    ) -> int:
        """
        Delete all samples between two timestamps for a given time series.

        :param key: Key name for the time series.
        :param fromtimestamp: Start timestamp for the range deletion.
        :param totimestamp: End timestamp for the range deletion.
        :return: The number of samples that were deleted, or an error reply.
        """
        return await self.execute_module_command(
            CommandName.TS_DEL,
            key,
            normalized_timestamp(fromtimestamp),
            normalized_timestamp(totimestamp),
            callback=IntCallback(),
        )

    @module_command(
        CommandName.TS_ALTER,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def alter(
        self,
        key: KeyT,
        labels: Optional[Mapping[StringT, StringT]] = None,
        retention: Optional[int] = None,
        chunk_size: Optional[int] = None,
        duplicate_policy: Optional[
            Literal[
                PureToken.BLOCK,
                PureToken.FIRST,
                PureToken.LAST,
                PureToken.MAX,
                PureToken.MIN,
                PureToken.SUM,
            ]
        ] = None,
    ) -> bool:
        """
        Update the retention, chunk size, duplicate policy, and labels of an existing time series.

        :param key: Key name for the time series.
        :param labels: Dictionary mapping labels to values that represent metadata labels of the
         key and serve as a secondary index.
        :param retention: Maximum retention period, compared to the maximum existing timestamp, in
         milliseconds.
        :param chunk_size: Initial allocation size, in bytes, for the data part of each new chunk.
        :param duplicate_policy: Policy for handling multiple samples with identical timestamps.
        :return: True if executed correctly, False otherwise.
        """
        pieces: CommandArgList = [key]
        if labels:
            pieces.extend(
                [
                    PrefixToken.LABELS,
                    *dict_to_flat_list(labels),  # type: ignore
                ]
            )
        if retention is not None:
            pieces.extend([PrefixToken.RETENTION, retention])
        if chunk_size is not None:
            pieces.extend([PrefixToken.CHUNK_SIZE, chunk_size])
        if duplicate_policy:
            pieces.extend([PrefixToken.DUPLICATE_POLICY, duplicate_policy])
        return await self.execute_module_command(
            CommandName.TS_ALTER, *pieces, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.TS_ADD,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def add(
        self,
        key: KeyT,
        timestamp: Union[int, datetime, StringT],
        value: Union[int, float],
        retention: Optional[int] = None,
        encoding: Optional[
            Literal[PureToken.COMPRESSED, PureToken.UNCOMPRESSED]
        ] = None,
        chunk_size: Optional[int] = None,
        duplicate_policy: Optional[
            Literal[
                PureToken.BLOCK,
                PureToken.FIRST,
                PureToken.LAST,
                PureToken.MAX,
                PureToken.MIN,
                PureToken.SUM,
            ]
        ] = None,
        labels: Optional[Mapping[StringT, ValueT]] = None,
    ) -> int:
        """
        Add a sample to a time series.

        :param key: Name of the time series.
        :param timestamp: UNIX sample timestamp in milliseconds or `*` to set the timestamp
         according to the server clock.
        :param value: Numeric data value of the sample.
        :param retention: Maximum retention period, compared to the maximum existing timestamp, in
         milliseconds.
        :param encoding: Encoding format for the series sample.
        :param chunk_size: Memory size, in bytes, allocated for each data chunk.
        :param duplicate_policy: Policy for handling samples with identical timestamps.
        :param labels: Dictionary of labels associated with the sample.
        :return: Number of samples added to the time series.
        """
        pieces: CommandArgList = [
            key,
            normalized_timestamp(timestamp),
            value,
        ]
        if retention is not None:
            pieces.extend([PrefixToken.RETENTION, retention])
        if encoding:
            pieces.extend([PrefixToken.ENCODING, encoding])
        if chunk_size is not None:
            pieces.extend([PrefixToken.CHUNK_SIZE, chunk_size])
        if duplicate_policy:
            pieces.extend([PrefixToken.ON_DUPLICATE, duplicate_policy])
        if labels:
            pieces.extend(
                [
                    PrefixToken.LABELS,
                    *dict_to_flat_list(labels),  # type: ignore
                ]
            )
        return await self.execute_module_command(
            CommandName.TS_ADD, *pieces, callback=IntCallback()
        )

    @module_command(
        CommandName.TS_MADD,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def madd(
        self, ktvs: Parameters[Tuple[AnyStr, int, Union[int, float]]]
    ) -> Tuple[int, ...]:
        """
        Append new samples to one or more time series.

        :param ktvs: A list of tuples, where each tuple contains the key name for the time series,
         an integer UNIX sample timestamp in milliseconds or `*` to set the timestamp according
         to the server clock, and a numeric data value of the sample.
        :return: A tuple of integers representing the timestamp of each added sample
        """
        pieces: CommandArgList = list(itertools.chain(*ktvs))

        return await self.execute_module_command(
            CommandName.TS_MADD, *pieces, callback=TupleCallback[int]()
        )

    @module_command(
        CommandName.TS_INCRBY,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def incrby(
        self,
        key: KeyT,
        value: Union[int, float],
        labels: Optional[Mapping[StringT, ValueT]] = None,
        timestamp: Optional[Union[datetime, int, StringT]] = None,
        retention: Optional[Union[int, timedelta]] = None,
        uncompressed: Optional[bool] = None,
        chunk_size: Optional[int] = None,
    ) -> int:
        """
        Increments the value of the sample with the maximum existing timestamp, or creates
        a new sample with a value equal to the value of the sample with the maximum existing
        timestamp with a given increment.

        :param key: Name of the time series.
        :param value: Numeric data value of the sample.
        :param labels: Set of label-value pairs that represent metadata labels of the key and serve
         as a secondary index. Use it only if you are creating a new time series.
        :param timestamp: UNIX sample timestamp in milliseconds or `*` to set the timestamp
         according to the server clock. `timestamp` must be equal to or higher than the maximum
         existing timestamp. When not specified, the timestamp is set according to the server clock.
        :param retention: Maximum retention period, compared to the maximum existing timestamp,
         in milliseconds. Use it only if you are creating a new time series.
        :param uncompressed: Changes data storage from compressed (default) to uncompressed.
         Use it only if you are creating a new time series.
        :param chunk_size: Memory size, in bytes, allocated for each data chunk.
         Use it only if you are creating a new time series.
        :return: The timestamp of the upserted sample, or an error.
        """
        pieces: CommandArgList = [key, value]
        if timestamp:
            pieces.extend([PrefixToken.TIMESTAMP, normalized_timestamp(timestamp)])
        if retention:
            pieces.extend([PrefixToken.RETENTION, normalized_milliseconds(retention)])
        if uncompressed:
            pieces.append(PureToken.UNCOMPRESSED)
        if chunk_size:
            pieces.extend([PrefixToken.CHUNK_SIZE, chunk_size])
        if labels:
            pieces.extend(
                [PrefixToken.LABELS, *dict_to_flat_list(labels)]  # type: ignore
            )

        return await self.execute_module_command(
            CommandName.TS_INCRBY, *pieces, callback=IntCallback()
        )

    @module_command(
        CommandName.TS_DECRBY,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def decrby(
        self,
        key: KeyT,
        value: Union[int, float],
        labels: Optional[Mapping[StringT, ValueT]] = None,
        timestamp: Optional[Union[datetime, int, StringT]] = None,
        retention: Optional[Union[int, timedelta]] = None,
        uncompressed: Optional[bool] = None,
        chunk_size: Optional[int] = None,
    ) -> int:
        """
        Decrease the value of the sample with the maximum existing timestamp, or create a new
        sample with a value equal to the value of the sample with the maximum existing timestamp
        with a given decrement.

        :param key: Key name for the time series.
        :param value: Numeric data value of the sample.
        :param labels: Mapping of labels to values that represent metadata labels of the key
         and serve as a secondary index. Use it only if you are creating a new time series.
        :param timestamp: UNIX sample timestamp in milliseconds or `*` to set the timestamp
         according to the server clock. When not specified, the timestamp is set according
         to the server clock.
        :param retention: Maximum retention period, compared to the maximum existing timestamp,
         in milliseconds. Use it only if you are creating a new time series. It is ignored if
         you are adding samples to an existing time series.
        :param uncompressed: Changes data storage from compressed (default) to uncompressed.
         Use it only if you are creating a new time series. It is ignored if you are adding samples
         to an existing time series.
        :param chunk_size: Memory size, in bytes, allocated for each data chunk. Use it only if
         you are creating a new time series. It is ignored if you are adding samples to an existing
         time series.
        :return: The timestamp of the upserted sample, or an error if the operation failed.
        """
        pieces: CommandArgList = [key, value]

        if timestamp:
            pieces.extend([PrefixToken.TIMESTAMP, normalized_timestamp(timestamp)])
        if retention:
            pieces.extend([PrefixToken.RETENTION, normalized_milliseconds(retention)])
        if uncompressed:
            pieces.append(PureToken.UNCOMPRESSED)
        if chunk_size:
            pieces.extend([PrefixToken.CHUNK_SIZE, chunk_size])
        if labels:
            pieces.extend(
                [PrefixToken.LABELS, *dict_to_flat_list(labels)]  # type: ignore
            )
        return await self.execute_module_command(
            CommandName.TS_DECRBY, *pieces, callback=IntCallback()
        )

    @module_command(
        CommandName.TS_CREATERULE,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        arguments={"aligntimestamp": {"version_introduced": "1.8.0"}},
        module=MODULE,
    )
    async def createrule(
        self,
        source: KeyT,
        destination: KeyT,
        aggregation: Literal[
            PureToken.AVG,
            PureToken.COUNT,
            PureToken.FIRST,
            PureToken.LAST,
            PureToken.MAX,
            PureToken.MIN,
            PureToken.RANGE,
            PureToken.STD_P,
            PureToken.STD_S,
            PureToken.SUM,
            PureToken.TWA,
            PureToken.VAR_P,
            PureToken.VAR_S,
        ],
        bucketduration: Union[int, timedelta],
        aligntimestamp: Optional[int] = None,
    ) -> bool:
        """
        Create a compaction rule

        :param source: Key name for the source time series.
        :param destination: Key name for the destination (compacted) time series.
        :param aggregation: Aggregates results into time buckets by the given aggregation type
        :param bucketduration: Duration of each bucket, in milliseconds.
        :param aligntimestamp: Ensures that there is a bucket that starts exactly at
         ``aligntimestamp`` and aligns all other buckets accordingly. It is expressed
         in milliseconds. The default value is 0 aligned with the epoch.
        :return: True if executed correctly, False otherwise.
        """
        pieces: CommandArgList = [source, destination]
        pieces.extend(
            [
                PrefixToken.AGGREGATION,
                aggregation,
                normalized_milliseconds(bucketduration),
            ]
        )
        if aligntimestamp is not None:
            pieces.append(aligntimestamp)
        return await self.execute_module_command(
            CommandName.TS_CREATERULE, *pieces, callback=SimpleStringCallback()
        )

    @module_command(
        CommandName.TS_DELETERULE,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def deleterule(self, source: KeyT, destination: KeyT) -> bool:
        """
        Delete a compaction rule from a RedisTimeSeries sourceKey to a destinationKey.

        :param source: Key name for the source time series.
        :param destination: Key name for the destination (compacted) time series.
        :return: True if the command executed correctly, False otherwise.

        .. warning:: This command does not delete the compacted series.
        """
        pieces: CommandArgList = [source, destination]

        return await self.execute_module_command(
            CommandName.TS_DELETERULE, *pieces, callback=SimpleStringCallback()
        )

    @mutually_inclusive_parameters("min_value", "max_value")
    @mutually_inclusive_parameters("aggregator", "bucketduration")
    @module_command(
        CommandName.TS_RANGE,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        arguments={
            "latest": {"version_introduced": "1.8.0"},
            "empty": {"version_introduced": "1.8.0"},
        },
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def range(
        self,
        key: KeyT,
        fromtimestamp: Union[datetime, int, StringT],
        totimestamp: Union[datetime, int, StringT],
        *,
        filter_by_ts: Optional[Parameters[int]] = None,
        min_value: Optional[Union[int, float]] = None,
        max_value: Optional[Union[int, float]] = None,
        count: Optional[int] = None,
        aggregator: Optional[
            Literal[
                PureToken.AVG,
                PureToken.COUNT,
                PureToken.FIRST,
                PureToken.LAST,
                PureToken.MAX,
                PureToken.MIN,
                PureToken.RANGE,
                PureToken.STD_P,
                PureToken.STD_S,
                PureToken.SUM,
                PureToken.TWA,
                PureToken.VAR_P,
                PureToken.VAR_S,
            ]
        ] = None,
        bucketduration: Optional[Union[int, timedelta]] = None,
        align: Optional[Union[int, StringT]] = None,
        buckettimestamp: Optional[StringT] = None,
        empty: Optional[bool] = None,
        latest: Optional[bool] = None,
    ) -> Union[Tuple[Tuple[int, float], ...], Tuple[()]]:
        """
        Query a range in forward direction.

        :param key: The key name for the time series.
        :param fromtimestamp: Start timestamp for the range query (integer UNIX timestamp in
         milliseconds) or `-` to denote the timestamp of the earliest sample in the time series.
        :param totimestamp: End timestamp for the range query (integer UNIX timestamp in
         milliseconds) or `+` to denote the timestamp of the latest sample in the time series.
        :param filter_by_ts: List of specific timestamps to filter samples by.
        :param min_value: Minimum value to filter samples by.
        :param max_value: Maximum value to filter samples by.
        :param count: Limits the number of returned samples.
        :param aggregator: Aggregates samples into time buckets by the provided aggregation type.
        :param bucketduration: Duration of each bucket in milliseconds.
        :param align: Time bucket alignment control for :paramref:`aggregator`.
        :param buckettimestamp: Timestamp of the first bucket.
        :param empty: If True, returns an empty list instead of raising an error when no data
         is available.
        :param latest: Used when a time series is a compaction. When ``True``, the command also
         reports the compacted value of the latest, possibly partial, bucket, given that
         this bucket's start time falls within ``[fromtimestamp, totimestamp]``.

        :return: A tuple of samples, where each sample is a tuple of timestamp and value.
        """
        pieces: CommandArgList = [
            key,
            normalized_timestamp(fromtimestamp),
            normalized_timestamp(totimestamp),
        ]
        if latest:
            pieces.append(b"LATEST")
        if filter_by_ts:
            _ts: List[int] = list(filter_by_ts)
            pieces.extend([PrefixToken.FILTER_BY_TS, *_ts])
        if min_value is not None and max_value is not None:
            pieces.extend([PureToken.FILTER_BY_VALUE, min_value, max_value])
        if count is not None:
            pieces.extend([PrefixToken.COUNT, count])
        if aggregator and bucketduration is not None:
            if align is not None:
                pieces.extend([PrefixToken.ALIGN, align])
            pieces.extend(
                [
                    PrefixToken.AGGREGATION,
                    aggregator,
                    normalized_milliseconds(bucketduration),
                ]
            )
            if buckettimestamp is not None:
                pieces.extend([PureToken.BUCKETTIMESTAMP, buckettimestamp])
            if empty is not None:
                pieces.append(PureToken.EMPTY)

        return await self.execute_module_command(
            CommandName.TS_RANGE, *pieces, callback=SamplesCallback()
        )

    @mutually_inclusive_parameters("min_value", "max_value")
    @mutually_inclusive_parameters("aggregator", "bucketduration")
    @module_command(
        CommandName.TS_REVRANGE,
        group=COMMAND_GROUP,
        version_introduced="1.4.0",
        arguments={
            "latest": {"version_introduced": "1.8.0"},
            "empty": {"version_introduced": "1.8.0"},
        },
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def revrange(
        self,
        key: KeyT,
        fromtimestamp: Union[int, datetime, StringT],
        totimestamp: Union[int, datetime, StringT],
        *,
        filter_by_ts: Optional[Parameters[int]] = None,
        min_value: Optional[Union[int, float]] = None,
        max_value: Optional[Union[int, float]] = None,
        count: Optional[int] = None,
        aggregator: Optional[
            Literal[
                PureToken.AVG,
                PureToken.COUNT,
                PureToken.FIRST,
                PureToken.LAST,
                PureToken.MAX,
                PureToken.MIN,
                PureToken.RANGE,
                PureToken.STD_P,
                PureToken.STD_S,
                PureToken.SUM,
                PureToken.TWA,
                PureToken.VAR_P,
                PureToken.VAR_S,
            ]
        ] = None,
        bucketduration: Optional[Union[int, timedelta]] = None,
        align: Optional[Union[int, StringT]] = None,
        buckettimestamp: Optional[StringT] = None,
        empty: Optional[bool] = None,
        latest: Optional[bool] = None,
    ) -> Union[Tuple[Tuple[int, float], ...], Tuple[()]]:
        """
        Query a range in reverse direction from a RedisTimeSeries key.

        :param key: The key name for the time series.
        :param fromtimestamp: Start timestamp for the range query (integer UNIX timestamp
         in milliseconds) or `-` to denote the timestamp of the earliest sample in the time series.
        :param totimestamp: End timestamp for the range query (integer UNIX timestamp in
         milliseconds) or `+` to denote the timestamp of the latest sample in the time series.
        :param filter_by_ts: List of specific timestamps to filter samples by.
        :param min_value: Minimum value to filter samples by.
        :param max_value: Maximum value to filter samples by.
        :param count: Limit the number of returned samples.
        :param aggregator: Aggregates samples into time buckets by the provided aggregation type.
        :param bucketduration: Duration of each bucket in milliseconds.
        :param align: Time bucket alignment control for :paramref:`aggregator`.
        :param buckettimestamp: Timestamp for the first bucket.
        :param empty: Return an empty list if no samples are found.
        :param latest: Report the compacted value of the latest, possibly partial, bucket.

        :return: A tuple of timestamp-value pairs in reverse order.
        """
        pieces: CommandArgList = [
            key,
            normalized_timestamp(fromtimestamp),
            normalized_timestamp(totimestamp),
        ]
        if latest:
            pieces.append(b"LATEST")
        if filter_by_ts:
            _ts: List[int] = list(filter_by_ts)
            pieces.extend([PrefixToken.FILTER_BY_TS, *_ts])
        if min_value is not None and max_value is not None:
            pieces.extend([PureToken.FILTER_BY_VALUE, min_value, max_value])
        if count is not None:
            pieces.extend([PrefixToken.COUNT, count])
        if aggregator and bucketduration is not None:
            if align is not None:
                pieces.extend([PrefixToken.ALIGN, align])
            pieces.extend(
                [
                    PrefixToken.AGGREGATION,
                    aggregator,
                    normalized_milliseconds(bucketduration),
                ]
            )
            if buckettimestamp is not None:
                pieces.extend([PureToken.BUCKETTIMESTAMP, buckettimestamp])
            if empty is not None:
                pieces.append(PureToken.EMPTY)

        return await self.execute_module_command(
            CommandName.TS_REVRANGE, *pieces, callback=SamplesCallback()
        )

    @mutually_inclusive_parameters("min_value", "max_value")
    @mutually_exclusive_parameters("withlabels", "selected_labels")
    @mutually_inclusive_parameters("aggregator", "bucketduration")
    @mutually_inclusive_parameters("groupby", "reducer")
    @module_command(
        CommandName.TS_MRANGE,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        arguments={
            "latest": {"version_introduced": "1.8.0"},
            "empty": {"version_introduced": "1.8.0"},
        },
        module=MODULE,
        cluster=ClusterCommandConfig(
            route=NodeFlag.PRIMARIES,
            combine=ClusterMergeTimeSeries(),
        ),
        flags={CommandFlag.READONLY},
    )
    async def mrange(
        self,
        fromtimestamp: Union[int, datetime, StringT],
        totimestamp: Union[int, datetime, StringT],
        filters: Optional[Parameters[StringT]] = None,
        *,
        filter_by_ts: Optional[Parameters[int]] = None,
        min_value: Optional[Union[int, float]] = None,
        max_value: Optional[Union[int, float]] = None,
        withlabels: Optional[bool] = None,
        selected_labels: Optional[Parameters[StringT]] = None,
        count: Optional[int] = None,
        align: Optional[Union[int, StringT]] = None,
        aggregator: Optional[
            Literal[
                PureToken.AVG,
                PureToken.COUNT,
                PureToken.FIRST,
                PureToken.LAST,
                PureToken.MAX,
                PureToken.MIN,
                PureToken.RANGE,
                PureToken.STD_P,
                PureToken.STD_S,
                PureToken.SUM,
                PureToken.TWA,
                PureToken.VAR_P,
                PureToken.VAR_S,
            ]
        ] = None,
        bucketduration: Optional[Union[int, timedelta]] = None,
        buckettimestamp: Optional[StringT] = None,
        groupby: Optional[StringT] = None,
        reducer: Optional[
            Literal[
                PureToken.AVG,
                PureToken.COUNT,
                PureToken.FIRST,
                PureToken.LAST,
                PureToken.MAX,
                PureToken.MIN,
                PureToken.RANGE,
                PureToken.STD_P,
                PureToken.STD_S,
                PureToken.SUM,
                PureToken.VAR_P,
                PureToken.VAR_S,
            ]
        ] = None,
        empty: Optional[bool] = None,
        latest: Optional[bool] = None,
    ) -> Dict[
        AnyStr,
        Tuple[Dict[AnyStr, AnyStr], Union[Tuple[Tuple[int, float], ...], Tuple[()]]],
    ]:
        """
        Query a range across multiple time series by filters in forward direction.

        :param fromtimestamp: Start timestamp for the range query (integer UNIX timestamp
         in milliseconds) or `-` to denote the timestamp of the earliest sample amongst
         all time series that passes the filters.
        :param totimestamp: End timestamp for the range query (integer UNIX timestamp in
         milliseconds) or `+` to denote the timestamp of the latest sample amongst all
         time series that passes the filters
        :param filters: Filter expressions to apply to the time series.
        :param filter_by_ts: Timestamps to filter the time series by.
        :param min_value: Minimum value to filter the time series by.
        :param max_value: Maximum value to filter the time series by.
        :param withlabels: Whether to include labels in the response.
        :param selected_labels: Returns a subset of the label-value pairs that represent metadata
         labels of the time series. Use when a large number of labels exists per series, but only
         the values of some of the labels are required. If :paramref:`withlabels` or
         :paramref:`selected_labels` are not specified, by default, an empty mapping is reported
         as label-value pairs.
        :param count: Limit the number of samples returned.
        :param align: Time bucket alignment control for :paramref:`aggregator`.
        :param aggregator: Aggregates samples into time buckets by the provided aggregation type.
        :param bucketduration: Duration of each bucket, in milliseconds.
        :param buckettimestamp: Timestamp of the first bucket.
        :param groupby: Label to group the samples by
        :param reducer: Aggregation type to aggregate the results in each group
        :param empty: Optional boolean to include empty time series in the response.
        :param latest: Report the compacted value of the latest, possibly partial, bucket.

        :return: A dictionary containing the time series data.
        """
        pieces: CommandArgList = [
            normalized_timestamp(fromtimestamp),
            normalized_timestamp(totimestamp),
        ]
        if latest:
            pieces.append(b"LATEST")
        if filter_by_ts:
            _ts: List[int] = list(filter_by_ts)
            pieces.extend([PrefixToken.FILTER_BY_TS, *_ts])
        if min_value is not None and max_value is not None:
            pieces.extend([PureToken.FILTER_BY_VALUE, min_value, max_value])
        if withlabels:
            pieces.append(PureToken.WITHLABELS)
        if selected_labels:
            _labels: List[StringT] = list(selected_labels)
            pieces.extend([PureToken.SELECTED_LABELS, *_labels])
        if count is not None:
            pieces.extend([PrefixToken.COUNT, count])
        if aggregator or buckettimestamp is not None:
            if align is not None:
                pieces.extend([PrefixToken.ALIGN, align])
            if aggregator and bucketduration is not None:
                pieces.extend(
                    [
                        PrefixToken.AGGREGATION,
                        aggregator,
                        normalized_milliseconds(bucketduration),
                    ]
                )
            if buckettimestamp is not None:
                pieces.extend([PureToken.BUCKETTIMESTAMP, buckettimestamp])
            if empty:
                pieces.append(PureToken.EMPTY)
        if filters:
            _filters: List[StringT] = list(filters)
            pieces.extend([PrefixToken.FILTER, *_filters])
        if groupby and reducer:
            pieces.extend([PureToken.GROUPBY, groupby, b"REDUCE", reducer])
        return await self.execute_module_command(
            CommandName.TS_MRANGE,
            *pieces,
            callback=TimeSeriesMultiCallback[AnyStr](),
            grouped=groupby is not None,
        )

    @mutually_inclusive_parameters("min_value", "max_value")
    @mutually_exclusive_parameters("withlabels", "selected_labels")
    @mutually_inclusive_parameters("aggregator", "bucketduration")
    @mutually_inclusive_parameters("groupby", "reducer")
    @module_command(
        CommandName.TS_MREVRANGE,
        group=COMMAND_GROUP,
        version_introduced="1.4.0",
        arguments={
            "latest": {"version_introduced": "1.8.0"},
            "empty": {"version_introduced": "1.8.0"},
        },
        module=MODULE,
        cluster=ClusterCommandConfig(
            route=NodeFlag.PRIMARIES, combine=ClusterMergeTimeSeries()
        ),
        flags={CommandFlag.READONLY},
    )
    async def mrevrange(
        self,
        fromtimestamp: Union[int, datetime, StringT],
        totimestamp: Union[int, datetime, StringT],
        filters: Optional[Parameters[StringT]] = None,
        *,
        filter_by_ts: Optional[Parameters[int]] = None,
        min_value: Optional[Union[int, float]] = None,
        max_value: Optional[Union[int, float]] = None,
        withlabels: Optional[bool] = None,
        selected_labels: Optional[Parameters[StringT]] = None,
        count: Optional[int] = None,
        align: Optional[Union[int, StringT]] = None,
        aggregator: Optional[
            Literal[
                PureToken.AVG,
                PureToken.COUNT,
                PureToken.FIRST,
                PureToken.LAST,
                PureToken.MAX,
                PureToken.MIN,
                PureToken.RANGE,
                PureToken.STD_P,
                PureToken.STD_S,
                PureToken.SUM,
                PureToken.TWA,
                PureToken.VAR_P,
                PureToken.VAR_S,
            ]
        ] = None,
        bucketduration: Optional[Union[int, timedelta]] = None,
        buckettimestamp: Optional[StringT] = None,
        groupby: Optional[StringT] = None,
        reducer: Optional[StringT] = None,
        empty: Optional[bool] = None,
        latest: Optional[bool] = None,
    ) -> Dict[
        AnyStr,
        Tuple[Dict[AnyStr, AnyStr], Union[Tuple[Tuple[int, float], ...], Tuple[()]]],
    ]:
        """
        Query a range across multiple time series by filters in reverse direction.

        :param fromtimestamp: Start timestamp for the range query (integer UNIX timestamp
         in milliseconds) or `-` to denote the timestamp of the earliest sample amongst
         all time series that passes the filters.
        :param totimestamp: End timestamp for the range query (integer UNIX timestamp in
         milliseconds) or `+` to denote the timestamp of the latest sample amongst all
         time series that passes the filters
        :param filters: Filter expressions to apply to the time series.
        :param filter_by_ts: Timestamps to filter the time series by.
        :param min_value: Minimum value to filter the time series by.
        :param max_value: Maximum value to filter the time series by.
        :param withlabels: Whether to include labels in the response.
        :param selected_labels: Returns a subset of the label-value pairs that represent metadata
         labels of the time series. Use when a large number of labels exists per series, but only
         the values of some of the labels are required. If :paramref:`withlabels` or
         :paramref:`selected_labels` are not specified, by default, an empty mapping is reported
         as label-value pairs.
        :param count: Limit the number of samples returned.
        :param align: Time bucket alignment control for :paramref:`aggregator`.
        :param aggregator: Aggregates samples into time buckets by the provided aggregation type.
        :param bucketduration: Duration of each bucket, in milliseconds.
        :param buckettimestamp: Timestamp of the first bucket.
        :param groupby: Label to group the samples by
        :param reducer: Aggregation type to aggregate the results in each group
        :param empty: Optional boolean to include empty time series in the response.
        :param latest: Report the compacted value of the latest, possibly partial, bucket.

        :return: A dictionary containing the result of the query.
        """
        pieces: CommandArgList = [
            normalized_timestamp(fromtimestamp),
            normalized_timestamp(totimestamp),
        ]
        if latest:
            pieces.append(b"LATEST")
        if filter_by_ts:
            _ts: List[int] = list(filter_by_ts)
            pieces.extend([PrefixToken.FILTER_BY_TS, *_ts])
        if min_value is not None and max_value is not None:
            pieces.extend([PureToken.FILTER_BY_VALUE, min_value, max_value])
        if withlabels:
            pieces.append(PureToken.WITHLABELS)
        if selected_labels:
            _labels: List[StringT] = list(selected_labels)
            pieces.extend([PureToken.SELECTED_LABELS, *_labels])
        if count is not None:
            pieces.extend([PrefixToken.COUNT, count])
        if aggregator or buckettimestamp is not None:
            if align is not None:
                pieces.extend([PrefixToken.ALIGN, align])
            if aggregator and bucketduration is not None:
                pieces.extend(
                    [
                        PrefixToken.AGGREGATION,
                        aggregator,
                        normalized_milliseconds(bucketduration),
                    ]
                )
            if buckettimestamp is not None:
                pieces.extend([PureToken.BUCKETTIMESTAMP, buckettimestamp])
            if empty:
                pieces.append(PureToken.EMPTY)
        if filters:
            _filters: List[StringT] = list(filters)
            pieces.extend([PrefixToken.FILTER, *_filters])
        if groupby and reducer and reducer:
            pieces.extend([PureToken.GROUPBY, groupby, b"REDUCE", reducer])

        return await self.execute_module_command(
            CommandName.TS_MREVRANGE,
            *pieces,
            callback=TimeSeriesMultiCallback[AnyStr](),
            grouped=groupby is not None,
        )

    @module_command(
        CommandName.TS_GET,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        arguments={"latest": {"version_introduced": "1.8.0"}},
        module=MODULE,
        flags={CommandFlag.READONLY},
        cache_config=CacheConfig(lambda *a, **_: a[0]),
    )
    async def get(
        self, key: KeyT, latest: Optional[bool] = None
    ) -> Union[Tuple[int, float], Tuple[()]]:
        """
        Get the sample with the highest timestamp from a given time series.

        :param key: The key name for the time series.
        :param latest: If the time series is a compaction, if ``True``, reports
         the compacted value of the latest, possibly partial, bucket. When ``False``,
         does not report the latest, possibly partial, bucket. When a time series is not a
         compaction, the parameter is ignored.
        :return: A tuple of (timestamp, value) of the sample with the highest timestamp,
         or an empty tuple if the time series is empty.
        """
        pieces: CommandArgList = [key]
        if latest:
            pieces.append(b"LATEST")
        return await self.execute_module_command(
            CommandName.TS_GET, *pieces, callback=SampleCallback()
        )

    @mutually_exclusive_parameters("withlabels", "selected_labels")
    @module_command(
        CommandName.TS_MGET,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        arguments={"latest": {"version_introduced": "1.8.0"}},
        module=MODULE,
        cluster=ClusterCommandConfig(
            route=NodeFlag.PRIMARIES,
            combine=ClusterMergeTimeSeries(),
        ),
        flags={CommandFlag.READONLY},
    )
    async def mget(
        self,
        filters: Parameters[StringT],
        withlabels: Optional[bool] = None,
        selected_labels: Optional[Parameters[StringT]] = None,
        latest: Optional[bool] = None,
    ) -> Dict[AnyStr, Tuple[Dict[AnyStr, AnyStr], Union[Tuple[int, float], Tuple[()]]]]:
        """
        Get the sample with the highest timestamp from each time series matching a specific filter.

        :param filters: Filters time series based on their labels and label values. At least one
         `label=value` filter is required.
        :param withlabels: Includes in the reply all label-value pairs representing metadata labels
         of the time series. If :paramref:`withlabels` or :paramref:`selected_labels` are not
         specified, by default, an empty dictionary is reported as label-value pairs.
        :param selected_labels: Returns a subset of the label-value pairs that represent metadata
         labels of the time series. Use when a large number of labels exists per series, but only
         the values of some of the labels are required. If :paramref:`withlabels` or
         :paramref:`selected_labels` are not specified, by default, an empty mapping is reported
         as label-value pairs.
        :param latest: Used when a time series is a compaction. If ``True``, the command also
         reports the compacted value of the latest possibly partial bucket, given that this
         bucket's start time falls within `[fromTimestamp, toTimestamp]`. If ``False``,
         the command does not report the latest possibly partial bucket. When a time series is
         not a compaction, the argument is ignored
        :return: For each time series matching the specified filters, a dictionary is returned with
         the time series key name as the key and a tuple containing the label-value pairs and a
         single timestamp-value pair as the value.
        """
        pieces: CommandArgList = []
        if latest:
            pieces.append(b"LATEST")
        if withlabels:
            pieces.append(PureToken.WITHLABELS)
        if selected_labels:
            _labels: List[StringT] = list(selected_labels)
            pieces.extend([b"SELECTED_LABELS", *_labels])
        pieces.extend([PrefixToken.FILTER, *filters])
        return await self.execute_module_command(
            CommandName.TS_MGET, *pieces, callback=TimeSeriesCallback[AnyStr]()
        )

    @module_command(
        CommandName.TS_INFO,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
    )
    async def info(
        self, key: KeyT, debug: Optional[bool] = None
    ) -> Dict[AnyStr, ResponseType]:
        """
        Return information and statistics for a time series.

        :param key: Key name of the time series.
        :param debug: Optional flag to get a more detailed information about the chunks.
        :return: Dictionary with information about the time series (name-value pairs).
        """
        pieces: CommandArgList = [key]
        if debug:
            pieces.append(b"DEBUG")
        return await self.execute_module_command(
            CommandName.TS_INFO, *pieces, callback=TimeSeriesInfoCallback[AnyStr]()
        )

    @module_command(
        CommandName.TS_QUERYINDEX,
        group=COMMAND_GROUP,
        version_introduced="1.0.0",
        module=MODULE,
        cluster=ClusterCommandConfig(
            route=NodeFlag.PRIMARIES,
            combine=ClusterMergeSets(),
        ),
        flags={CommandFlag.READONLY},
    )
    async def queryindex(self, filters: Parameters[StringT]) -> Set[AnyStr]:
        """
        Get all time series keys matching a filter list.

        :param filters: A list of filter expressions to match time series based on their labels
         and label values. Each filter expression has one of the following syntaxes:

          - ``label=value``, where ``label`` equals ``value``
          - ``label!=value``, where ``label`` does not equal ``value``
          - ``label=``, where ``key`` does not have label ``label``
          - ``label!=``, where ``key`` has label ``label``
          - ``label=(value1,value2,...)``, where ``key`` with label ``label`` equals one of
            the values in the list
          - ``label!=(value1,value2,...)``, where key with label ``label`` does not equal
            any of the values in the list

         At least one ``label=value`` filter is required. Filters are conjunctive. For example, the
         filter ``type=temperature room=study`` means the a time series is a temperature time series
         of a study room. Don't use whitespaces in the filter expression.
        :return: A set of time series keys matching the filter list. The set is empty if no time
         series matches the filter. An error is returned on invalid filter expression.

        """
        pieces: CommandArgList = [*filters]

        return await self.execute_module_command(
            CommandName.TS_QUERYINDEX, *pieces, callback=SetCallback[AnyStr]()
        )



from __future__ import annotations

import dataclasses
import random
import warnings
from typing import TYPE_CHECKING, Any

from coredis._utils import b, hash_slot, nativestr
from coredis.exceptions import (
    ConnectionError,
    RedisClusterException,
    RedisError,
    ResponseError,
)
from coredis.typing import (
    Dict,
    Iterable,
    Iterator,
    List,
    Literal,
    Node,
    Optional,
    Set,
    StringT,
    ValueT,
)

HASH_SLOTS = 16384
HASH_SLOTS_SET = set(range(HASH_SLOTS))

if TYPE_CHECKING:
    from coredis import Redis


@dataclasses.dataclass
class ManagedNode:
    """
    Represents a cluster node (primary or replica) in a redis cluster
    """

    host: str
    port: int
    server_type: Optional[Literal["primary", "replica"]] = None
    node_id: Optional[str] = None

    @property
    def name(self) -> str:
        return f"{self.host}:{self.port}"


class NodeManager:
    """
    Utility class to manage the topology of a redis cluster
    """

    def __init__(
        self,
        startup_nodes: Optional[Iterable[Node]] = None,
        reinitialize_steps: Optional[int] = None,
        skip_full_coverage_check: bool = False,
        nodemanager_follow_cluster: bool = True,
        decode_responses: bool = False,
        **connection_kwargs: Optional[Any],
    ) -> None:
        """
        :skip_full_coverage_check:
            Skips the check of cluster-require-full-coverage config, useful for clusters
            without the CONFIG command (like aws)
        :nodemanager_follow_cluster:
            The node manager will during initialization try the last set of nodes that
            it was operating on. This will allow the client to drift along side the cluster
            if the cluster nodes move around a slot.
        """
        self.connection_kwargs = connection_kwargs
        self.connection_kwargs.update(decode_responses=decode_responses)

        self.nodes: Dict[str, ManagedNode] = {}
        self.slots: Dict[int, List[ManagedNode]] = {}
        self.startup_nodes: List[ManagedNode] = (
            []
            if startup_nodes is None
            else list(ManagedNode(n["host"], n["port"]) for n in startup_nodes if n)
        )
        self.startup_nodes_reachable = False
        self.orig_startup_nodes = list(self.startup_nodes)
        self.reinitialize_counter = 0
        self.reinitialize_steps = reinitialize_steps or 25
        self._skip_full_coverage_check = skip_full_coverage_check
        self.nodemanager_follow_cluster = nodemanager_follow_cluster
        self.replicas_per_shard = 0

    def keys_to_nodes_by_slot(
        self, *keys: ValueT
    ) -> Dict[str, Dict[int, List[ValueT]]]:
        mapping: Dict[str, Dict[int, List[ValueT]]] = {}
        for k in keys:
            node = self.node_from_slot(hash_slot(b(k)))
            if node:
                mapping.setdefault(node.name, {}).setdefault(
                    hash_slot(b(k)), []
                ).append(k)
        return mapping

    def node_from_slot(self, slot: int) -> Optional[ManagedNode]:
        for node in self.slots[slot]:
            if node.server_type == "primary":
                return node
        return None  # noqa

    def all_nodes(self) -> Iterator[ManagedNode]:
        yield from self.nodes.values()

    def all_primaries(self) -> Iterator[ManagedNode]:
        for node in self.nodes.values():
            if node.server_type == "primary":
                yield node

    def all_replicas(self) -> Iterator[ManagedNode]:
        for node in self.nodes.values():
            if node.server_type == "replica":
                yield node

    def random_startup_node_iter(self, primary: bool = False) -> Iterator[ManagedNode]:
        """A generator that returns a random startup nodes"""
        options = list(
            self.all_primaries()
            if primary
            else (self.startup_nodes if self.startup_nodes_reachable else [])
        )
        while options:
            choice = random.choice(options)
            options.remove(choice)
            yield choice

    def random_node(self, primary: bool = True) -> ManagedNode:
        if primary:
            return random.choice(list(self.all_primaries()))
        else:
            return random.choice(list(self.nodes.values()))

    def get_redis_link(self, host: str, port: int) -> Redis[Any]:
        from coredis.client import Redis

        allowed_keys = (
            "username",
            "password",
            "encoding",
            "decode_responses",
            "stream_timeout",
            "connect_timeout",
            "ssl_context",
            "parser_class",
            "loop",
            "protocol_version",
        )
        connection_kwargs = {
            k: v for k, v in self.connection_kwargs.items() if k in allowed_keys
        }
        return Redis(host=host, port=port, **connection_kwargs)  # type: ignore

    async def initialize(self) -> None:
        """
        Initializes the slots cache by asking all startup nodes what the
        current cluster configuration is.

        TODO: Currently the last node will have the last say about how the configuration is setup.
        Maybe it should stop to try after it have correctly covered all slots or when one node is
        reached and it could execute CLUSTER SLOTS command.
        """
        nodes_cache: Dict[str, ManagedNode] = {}
        tmp_slots: Dict[int, List[ManagedNode]] = {}

        all_slots_covered = False
        disagreements: List[str] = []
        self.startup_nodes_reachable = False

        nodes = self.orig_startup_nodes
        replicas: Set[str] = set()
        startup_node_errors: Dict[str, List[str]] = {}

        # With this option the client will attempt to connect to any of the previous set of nodes
        # instead of the original set of startup nodes
        if self.nodemanager_follow_cluster:
            nodes = self.startup_nodes

        for node in nodes:
            cluster_slots = {}
            try:
                if node:
                    r = self.get_redis_link(host=node.host, port=node.port)
                    cluster_slots = await r.cluster_slots()
                    self.startup_nodes_reachable = True
            except RedisError as err:
                startup_node_errors.setdefault(str(err), []).append(node.name)
                continue

            all_slots_covered = True
            # If there's only one server in the cluster, its ``host`` is ''
            # Fix it to the host in startup_nodes
            if len(cluster_slots) == 1 and len(self.startup_nodes) == 1:
                slots = cluster_slots.get((0, HASH_SLOTS - 1))
                assert slots
                single_node_slots = slots[0]
                if len(single_node_slots["host"]) == 0:
                    single_node_slots["host"] = self.startup_nodes[0].host
                    single_node_slots["server_type"] = "master"

            for min_slot, max_slot in cluster_slots:
                _nodes = cluster_slots.get((min_slot, max_slot))
                assert _nodes
                primary_node = ManagedNode(
                    host=_nodes[0]["host"],
                    port=_nodes[0]["port"],
                    server_type="primary",
                    node_id=_nodes[0]["node_id"],
                )
                replica_nodes = [
                    ManagedNode(
                        host=n["host"],
                        port=n["port"],
                        server_type="replica",
                        node_id=n["node_id"],
                    )
                    for n in _nodes[1:]
                ]

                primary_node.host = primary_node.host or node.host
                nodes_cache[primary_node.name] = primary_node

                for i in range(min_slot, max_slot + 1):
                    if i not in tmp_slots:
                        tmp_slots[i] = [primary_node]
                        for replica_node in replica_nodes:
                            nodes_cache[replica_node.name] = replica_node
                            tmp_slots[i].append(replica_node)
                            replicas.add(replica_node.name)
                    else:
                        # Validate that 2 nodes want to use the same slot cache setup
                        if tmp_slots[i][0].name != node.name:
                            disagreements.append(
                                "{} vs {} on slot: {}".format(
                                    tmp_slots[i][0].name, node.name, i
                                ),
                            )
                            if len(disagreements) > 5:
                                raise RedisClusterException(
                                    "startup_nodes could not agree on a valid slots cache."
                                    f" {', '.join(disagreements)}"
                                )

                self.refresh_table_asap = False

            if not self._skip_full_coverage_check and (
                await self.cluster_require_full_coverage(nodes_cache)
            ):
                all_slots_covered = set(tmp_slots.keys()) == HASH_SLOTS_SET

            if all_slots_covered:
                break

        if not self.startup_nodes_reachable:
            details = ""
            # collapse any startup nodes by error representation
            if startup_node_errors:
                details = " Underlying errors:\n" + "\n".join(
                    [
                        f"- {err} [{','.join(nodes)}]"
                        for err, nodes in startup_node_errors.items()
                    ]
                )
            raise RedisClusterException(
                "Redis Cluster cannot be connected. "
                "Please provide at least one reachable node."
                f"{details}"
            )

        if not all_slots_covered:
            raise RedisClusterException(
                "Not all slots are covered after query all startup_nodes. "
                "{} of {} covered...".format(len(tmp_slots), HASH_SLOTS)
            )

        # Set the tmp variables to the real variables
        self.slots = tmp_slots
        self.nodes = nodes_cache
        self.replicas_per_shard = int(
            (len(self.nodes) / len(replicas)) - 1 if replicas else 0
        )
        self.reinitialize_counter = 0
        self.populate_startup_nodes()

    async def increment_reinitialize_counter(self, ct: int = 1) -> None:
        for _ in range(min(ct, self.reinitialize_steps)):
            self.reinitialize_counter += 1
            if self.reinitialize_counter % self.reinitialize_steps == 0:
                await self.initialize()

    async def node_require_full_coverage(self, node: ManagedNode) -> bool:
        try:
            r_node = self.get_redis_link(host=node.host, port=node.port)
            node_config = await r_node.config_get(["cluster-require-full-coverage"])
            return "yes" in node_config.values()
        except ResponseError as err:
            warnings.warn(
                "Unable to determine whether the cluster requires full coverage "
                f"due to response error from `CONFIG GET`: {err}. To suppress this "
                "warning use skip_full_coverage=True when initializing the client."
            )
            return False

    async def cluster_require_full_coverage(
        self, nodes_cache: Dict[str, ManagedNode]
    ) -> bool:
        """
        If exists 'cluster-require-full-coverage no' config on redis servers,
        then even all slots are not covered, cluster still will be able to
        respond
        """
        nodes = nodes_cache or self.nodes

        for node in nodes.values():
            try:
                if await self.node_require_full_coverage(node):
                    return True
            except ConnectionError:
                continue
        return False

    def set_node(
        self,
        host: StringT,
        port: int,
        server_type: Literal["primary", "replica"],
    ) -> ManagedNode:
        """Updates data for a node"""
        node = ManagedNode(
            host=nativestr(host),
            port=port,
            server_type=server_type,
            node_id=None,
        )
        self.nodes[node.name] = node
        return node

    def populate_startup_nodes(self) -> None:
        self.startup_nodes.clear()
        for n in self.nodes.values():
            self.startup_nodes.append(n)

    async def reset(self) -> None:
        await self.initialize()
